/***********************************************************
 * $Id: OsgiExecutorServiceProvider.java 1111 2016-02-01 14:42:55Z util $
 * 
 * scheduler utilities of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 02.12.2013
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.sched.osgi;

import java.io.IOException;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.osgi.framework.Bundle;
import org.osgi.framework.ServiceRegistration;
import org.osgi.service.cm.Configuration;
import org.osgi.service.cm.ConfigurationAdmin;
import org.osgi.service.cm.ConfigurationEvent;
import org.osgi.service.cm.ConfigurationException;
import org.osgi.service.cm.ConfigurationListener;
import org.osgi.service.cm.ManagedService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>A provider of an {@link ExecutorService} and
 * a {@link ScheduledExecutorService} as OSGi service for
 * use by various bundles.</p>
 * 
 * <p>This provider reads the OSGi configuration with the PID
 * <code>org.clazzes.util.sched</code> and interprets the two configuration
 * properties <code>executorPoolSize</code> and <code>scheduledCorePoolSize</code>.
 * </p>
 * 
 * <p>Typically, the exported thread pools are used like follows:</p>
 * <pre>
 *   &lt;bp:reference id="executorService" availability="optional" interface="java.util.concurrent.ExecutorService"
 *              filter="(pool.name=sched-util)"&gt;
 *   &lt;/bp:reference&gt;
 *     &lt;bp:bean id="oneTimeScheduler" 
 *                 class="org.clazzes.util.sched.impl.OneTimeSchedulerImpl" init-method="start" 
 *                 destroy-method="shutdownNow"&gt;
 *     &lt;bp:property name="executorService" ref="executorService"/&gt;
 *     &lt;bp:property name="resultLifeTime" value="60000"/&gt;
 *     &lt;bp:property name="gcInterval" value="60000"/&gt;
 *     &lt;bp:property name="threadLocalValues"&gt;
 *       &lt;bp:map&gt;
 *         &lt;bp:entry key="org.clazzes.login::Locale"&gt;&lt;bp:null/&gt;&lt;/bp:entry&gt;
 *         &lt;bp:entry key="org.clazzes.login::Principal"&gt;&lt;bp:null/&gt;&lt;/bp:entry&gt;
 *       &lt;/bp:map&gt;
 *     &lt;/bp:property&gt;
 *   &lt;/bp:bean&gt;
 * </pre>
 * <p>An example involving the exported scheduled executor service is:</p>
 * <pre>
 *   &lt;bp:reference id="executorService" availability="optional" interface="java.util.concurrent.ScheduledExecutorService"
 *              filter="(pool.name=sched-util)"&gt;
 *   &lt;/bp:reference&gt;
 *     &lt;bp:bean id="oneTimeScheduler" 
 *                 class="org.clazzes.util.sched.impl.OneTimeSchedulerImpl" init-method="start" 
 *                 destroy-method="shutdownNow"&gt;
 *     &lt;bp:property name="executorService" ref="executorService"/&gt;
 *     &lt;bp:property name="resultLifeTime" value="60000"/&gt;
 *     &lt;bp:property name="gcInterval" value="60000"/&gt;
 *     &lt;bp:property name="threadLocalValues"&gt;
 *       &lt;bp:map&gt;
 *         &lt;bp:entry key="org.clazzes.login::Locale"&gt;&lt;bp:null/&gt;&lt;/bp:entry&gt;
 *         &lt;bp:entry key="org.clazzes.login::Principal"&gt;&lt;bp:null/&gt;&lt;/bp:entry&gt;
 *       &lt;/bp:map&gt;
 *     &lt;/bp:property&gt;
 *   &lt;/bp:bean&gt;
 * </pre>
 */
public class OsgiExecutorServiceProvider implements ManagedService, ConfigurationListener {

    private static final Logger log = LoggerFactory.getLogger(OsgiExecutorServiceProvider.class);
    
    private static final int DEFAULT_EXECUTOR_POOL_SIZE = 4;
    private static final int DEFAULT_SCHEDULED_CORE_POOL_SIZE = 1;
    
    private static final String CONFIG_PID = "org.clazzes.util.sched";
    
    private Bundle bundle;
    private ConfigurationAdmin configurationAdmin;
    
    private int executorPoolSize;
    private int scheduledCorePoolSize;
    
    private ExecutorService executor;
    private ServiceRegistration<ExecutorService> executorRegistration;
    private ScheduledExecutorService scheduledExecutor;
    private ServiceRegistration<ScheduledExecutorService> scheduledExecutorRegistration;

    private final Dictionary<String,String> properties;
    
    public  OsgiExecutorServiceProvider() {
        this.executorPoolSize = DEFAULT_EXECUTOR_POOL_SIZE;
        this.scheduledCorePoolSize = DEFAULT_SCHEDULED_CORE_POOL_SIZE;
        
        this.properties = new Hashtable<String, String>();
        
        this.properties.put("pool.name","sched-util");
    }
    
    private static Dictionary<String, Object> getDefaultConfig() {
        
        Dictionary<String,Object> props = new Hashtable<String,Object>();
        props.put("executorPoolSize",String.valueOf(DEFAULT_EXECUTOR_POOL_SIZE));
        props.put("scheduledCorePoolSize",String.valueOf(DEFAULT_SCHEDULED_CORE_POOL_SIZE));
        return props;
    }
    
    public void createDefaultConfig() {
        
        try {
            Configuration cfg = this.configurationAdmin.getConfiguration(CONFIG_PID);
            
            Dictionary<String, Object> props = cfg.getProperties();
            
            if (props == null) {
                
                props = getDefaultConfig();
                
                log.info ("Creating default configuration [{}] for PID [{}].",props,CONFIG_PID);
                
                cfg.update(props);
            }

            this.updated(props);
            
        } catch (IOException e) {
            log.warn("Caught I/O exception while inspecting configuration PID ["+CONFIG_PID+"]",e);
        } catch (ConfigurationException e) {
            log.warn("Caught exception while activating configuration from PID ["+CONFIG_PID+"]",e);
        }
    }
    
    @Override
    public void configurationEvent(ConfigurationEvent event) {
        
        if (CONFIG_PID.equals(event.getPid()) &&
                event.getType() == ConfigurationEvent.CM_UPDATED) {
            
            Dictionary<String, Object> props;
            
            try {
                props = this.configurationAdmin.getConfiguration(CONFIG_PID).getProperties();
                
            } catch (IOException e) {
                
                log.warn("Caught I/O exception while reading configuration PID ["+CONFIG_PID+"] upon update, using default config",e);
                props = getDefaultConfig();
            }

            try {
                this.updated(props);
            } catch (ConfigurationException e) {
                log.warn("Caught exception while activating configuration from PID ["+CONFIG_PID+"]",e);
            }
        }
        
    }

    @Override
    public void updated(Dictionary<String,?> properties) throws ConfigurationException {
        
        if (log.isDebugEnabled()) {
            log.debug("updated called with properties [{}].",properties);
        }
        
        if (properties ==  null) {
            return;
        }
        
        int newExecutorPoolSize = this.executorPoolSize;
        int newScheduledCorePoolSize = this.scheduledCorePoolSize;

        Object executorPoolSize_o = properties.get("executorPoolSize");

        if (executorPoolSize_o != null) {
            try {

                newExecutorPoolSize = Integer.parseInt(executorPoolSize_o.toString());

                if (log.isDebugEnabled()) {
                    log.debug("Parameter executorPoolSize parsed to [{}].",newExecutorPoolSize);
                }

                if (newExecutorPoolSize < 1) {

                    newExecutorPoolSize = DEFAULT_EXECUTOR_POOL_SIZE;

                    log.error("Number [{}] less than one given for option executorPoolSize, using default value [{}] instead.",
                            executorPoolSize_o,newExecutorPoolSize);
                }

            }
            catch(NumberFormatException e) {

                newExecutorPoolSize = DEFAULT_EXECUTOR_POOL_SIZE;

                log.error("Non-number [{}] given for option executorPoolSize, using default value [{}] instead.",
                        executorPoolSize_o,newExecutorPoolSize);
            }
        }

        Object scheduledCorePoolSize_o = properties.get("scheduledCorePoolSize");

        if (scheduledCorePoolSize_o != null) {
            try {

                newScheduledCorePoolSize = Integer.parseInt(scheduledCorePoolSize_o.toString());

                if (log.isDebugEnabled()) {
                    log.debug("Parameter scheduledCorePoolSize parsed to [{}].",newScheduledCorePoolSize);
                }

                if (newScheduledCorePoolSize < 1) {

                    newScheduledCorePoolSize= DEFAULT_EXECUTOR_POOL_SIZE;

                    log.error("Number [{}] less than one given for option scheduledCorePoolSize, using default value [{}] instead.",
                            scheduledCorePoolSize_o,newScheduledCorePoolSize);
                }

            }
            catch(NumberFormatException e) {

                newScheduledCorePoolSize = DEFAULT_EXECUTOR_POOL_SIZE;

                log.error("Non-number [{}] given for option scheduledCorePoolSize, using default value [{}] instead.",
                        scheduledCorePoolSize_o,newScheduledCorePoolSize);
            }
        }
        
        synchronized(this) {
        
            if (newExecutorPoolSize != this.executorPoolSize) {
                
                this.executorPoolSize = newExecutorPoolSize;
                this.shutdownInternals(false);
            }
            
            if (this.executor == null) {
                
                log.info("Starting executor service with pool size [{}].",this.executorPoolSize);
                
                this.startInternals();
            }
            
            if (newScheduledCorePoolSize != this.scheduledCorePoolSize) {
                this.shutdownScheduledInternals(false);
            }
            
            if (this.scheduledExecutor == null) {
                
                log.info("Starting scheduled executor service with core pool size [{}].",this.scheduledCorePoolSize);

                this.startScheduledInternals();
            }
        }
    }
    
    @SuppressWarnings("unchecked")
    protected void startInternals() {
        
        if (this.executor == null) {
        
            this.executor = Executors.newFixedThreadPool(this.executorPoolSize);
            
            this.executorRegistration =
                    (ServiceRegistration<ExecutorService>)
                    this.bundle.getBundleContext().registerService(ExecutorService.class.getName(),this.executor,this.properties);
        }
    }

    protected ExecutorService shutdownInternals(boolean hard) {
        
        if (this.executorRegistration != null) {
            this.executorRegistration.unregister();
            this.executorRegistration = null;
        }
        
        ExecutorService ret = this.executor;

        if (this.executor != null) {
            
            if (hard) {
                
                log.info("Shutting down executor service upon bundle stop.");
                this.executor.shutdownNow();
            }
            else {
                log.info("Shutting down executor service upon change of configuration.");
                this.executor.shutdown();
            }
            this.executor = null;
        }
        return ret;
    }
    
    @SuppressWarnings("unchecked")
    protected void startScheduledInternals() {
        
        if (this.scheduledExecutor == null) {
        
            this.scheduledExecutor = Executors.newScheduledThreadPool(this.scheduledCorePoolSize);
            
            this.scheduledExecutorRegistration =
                    (ServiceRegistration<ScheduledExecutorService>)
                    this.bundle.getBundleContext().registerService(ScheduledExecutorService.class.getName(),this.scheduledExecutor,this.properties);
        }
    }

    protected ExecutorService shutdownScheduledInternals(boolean hard) {

        if (this.scheduledExecutorRegistration != null) {
            this.scheduledExecutorRegistration.unregister();
            this.scheduledExecutorRegistration = null;
        }
        
        ExecutorService ret = this.scheduledExecutor;

        if (this.scheduledExecutor != null) {
            
            if (hard) {
                
                log.info("Shutting down scheduled executor service upon bundle stop.");
                this.scheduledExecutor.shutdownNow();
            }
            else {
                log.info("Shutting down scheduled executor service upon change of configuration.");
                this.scheduledExecutor.shutdown();
            }
            this.scheduledExecutor = null;
        }
        
        return ret;
    }
    
    public void destroy() {
        
        ExecutorService svc1;
        ExecutorService svc2;
         
        synchronized (this) {
            
            svc1 = this.shutdownInternals(true);
            svc2 = this.shutdownScheduledInternals(true);
        }
        
        int iwait = 0;
        
        try {
            do {
                
                if (svc1 != null) {
                    
                    if (svc1.isTerminated()) {
                        log.info("All jobs of executor service have finished upon bundle stop.");
                        svc1 = null;
                    }
                    else {
                        
                        log.info("Waiting for all jobs of executor service to finish upon bundle stop...");
                        ++iwait;
                        
                        if (svc1.awaitTermination(1000,TimeUnit.MILLISECONDS)) {

                            log.info("All jobs of executor service have finished upon bundle stop.");
                            svc1 = null;
                        }
                        else {
                            log.warn("Not all jobs of executor service have finished upon bundle stop, will wait another second...");
                        }
                    }
                }
                
                if (svc2 != null) {
                    
                    if (svc2.isTerminated()) {
                        log.info("All jobs of scheduled executor service have finished upon bundle stop.");
                        svc2 = null;
                    }
                    else {
                        
                        log.info("Waiting for all jobs of scheduled executor service to finish upon bundle stop...");
                        ++iwait;

                        if (svc2.awaitTermination(1,TimeUnit.MILLISECONDS)) {
                            log.info("All jobs of scheduled executor service have finished upon bundle stop.");
                            svc2 = null;
                        }
                        else {
                            log.warn("Not all jobs of scheduled executor service have finished upon bundle stop, will wait another second...");
                        }
                    }
                }
            }
            while ((svc1 != null || svc2 != null) && iwait < 30);
            
            if (svc1 == null && svc2 == null) {
                log.info("All scheduler services have successfully been shut down upon bundle stop.");
            }
            else {
                log.warn("Not all scheduler services could be shut down upon bundle stop, scheduler threads may be pending.");                
            }

        } catch (InterruptedException e) {
            log.warn("Wait for termination of all scheduled jobs upon bundle stop has been interrupted.");
        }
        
    }

     public void setBundle(Bundle bundle) {
        this.bundle = bundle;
     }

    public void setConfigurationAdmin(ConfigurationAdmin configurationAdmin) {
        this.configurationAdmin = configurationAdmin;
    }
     
}
